"""
Manage runit services.

Options:

    * command  : str, None
               : service command to run, any of:
               : - start
               : - stop
               : - restart

    * target   : str, None
               : servcie directory to operate

    * wait_sec : int, None
               : wait for status changes before exit

    * force    : bool, False
               : forceful execution of service command

Requirements:

    * runit
      to install, for example run `apt-get install runit`
"""
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        c = cfg.get('wait_sec', None)
        if c is not None:
            args.append('-w%s' % c)

        command = cfg.get('command', None)
        if command is None:
            self.bld.fatal('command option is required.')

        c = cfg.get('force', False)
        if c:
            args.append('force-' + command)
        else:
            args.append(command)

        target = cfg.get('target', None)
        if target:
            target = expand_resource(self.group, target)
        if target is None:
            self.bld.fatal('target option is required.')
        args.append(target)


    def perform(self):
        if self.file_in:
            self.bld.fatal('%s takes no input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command([executable] + self.args)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program('sv')[0]



"""
Extract compressed files in a ZIP archive

Options:

    * extract_dir  : str, None,  destination directory (required)
    * quiet        : bool, False, quiet mode

Requirements:

    * unzip
      to install, for example run `apt-get install unzip`

"""

import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        # Update files, create if necessary
        self.args = ['-uo']

        c = self.conf.get('extract_dir')
        if c is None:
            self.bld.fatal('"extract_dir" configuration is required')

        path = c.format(**self.group.get_patterns())
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            os.remove(path)
            os.makedirs(path)
        self.args.append('-d ' + expand_resource(self.group, c))

        c = self.conf.get('quiet')
        if c:
            self.args.append('-q')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal('%s does not need output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
        ))


def configure(conf):
    bin_path = conf.find_program('unzip')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



""" Copy files.
"""
from shutil import copyfile, Error
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        try:
            copyfile(self.file_in[0], self.file_out[0])
            return 0
        except Error:
            self.bld.fatal('tried to copy file to itself')
        except IOError:
            self.bld.fatal('destination location cannot be written')
        return 1



"""
A mighty, modern linter that helps you avoid errors and enforce conventions in
your styles.

https://github.com/stylelint/stylelint


Options:

    * config : str, None
             : Path to a specific configuration file (JSON, YAML, or CommonJS),
             : or the name of a module in node_modules that points to one. If
             : no --config argument is provided, stylelint will search for
             : configuration files in the following places, in this order:
             :   - a stylelint property in package.json
             :   - a .stylelintrc file (with or without filename extension:
             :     .json, .yaml, .yml, and .js are available)
             :   - a stylelint.config.js file exporting a JS object
             : The search will begin in the working directory and move up the
             : directory tree until a configuration file is found.

    * config_basedir : str, None
                     : An absolute path to the directory that relative paths
                     : defining "extends" and "plugins" are *relative to*.
                     : Only necessary if these values are relative paths.

    * ignore_path : str, None
                  : Path to a file containing patterns that describe files to
                  : ignore. The path can be absolute or relative to
                  : process.cwd(). By default, stylelint looks for
                  : .stylelintignore in process.cwd().

    * ignore_pattern : str, None
                     : Pattern of files to ignore (in addition to those in
                     : .stylelintignore)

    * syntax : str, None
             : Specify a non-standard syntax. Options: "scss", "sass", "less",
             : "sugarss".
             : If you do not specify a syntax, non-standard syntaxes will be
             : automatically inferred by the file extensions .scss, .sass,
             : .less, and .sss.

    * fix : bool, None
          : Automatically fix violations of certain rules.

    * custom_syntax : str, None
                    : Module name or path to a JS file exporting a
                    : PostCSS-compatible syntax.

    * ignore_disables : bool, None
                      : Ignore styleline-disable comments.

    * disable_default_ignores : bool, None
                              : Allow linting of node_modules and
                              : bower_components.

    * cache : bool, None
            : Store the info about processed files in order to only operate on
            : the changed ones the next time you run stylelint. By default, the
            : cache is stored in "./.stylelintcache". To adjust this, use
            : --cache-location.

    * cache_location : str, None
                     : Path to a file or directory to be used for the cache
                     : location.
                     : Default is "./.stylelintcache". If a directory is
                     : specified, a cache file will be created inside the
                     : specified folder, with a name derived from a hash of the
                     : current working directory.
                     :
                     : If the directory for the cache does not exist, make sure
                     : you add a trailing "/" on *nix systems or "\\" on
                     : Windows. Otherwise the path will be assumed to be a file.

    * formatter : str, None
                : The output formatter:
                : ${getFormatterOptionsText({ useOr: true })}.

    * custom_formatter : str, None
                       : Path to a JS file exporting a custom formatting
                       : function.

    * quiet : bool, None
            : Only register warnings for rules with an "error"-level severity
            : (ignore "warning"-level).

    * color    : bool, None
    * no_color : bool, None
               : Force enabling/disabling of color.

    * report_needless_disables : bool, None
                               : Report stylelint-disable comments that are not
                               : blocking a lint warning.
                               : The process will exit with code
                               : ${EXIT_CODE_ERROR} if needless disables are
                               : found.

    * max_warnings : int, None
                   : Number of warnings above which the process will exit with
                   : code ${EXIT_CODE_ERROR}.
                   : Useful when setting "defaultSeverity" to "warning" and
                   : expecting the process to fail on warnings (e.g. CI build).


Requirements:

    * node.js
    * stylelint
      to install, run `npm install --save-dev stylelint`

"""

import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Specify a configuration file
        c = cfg.get('config_file')
        if c:
            args.append("--config='%s'" % expand_resource(self.group, c))

        self.add_bool_args('fix', 'ignore_disables', 'disable_default_ignores',
                'cache', 'quiet', 'color', 'no_color',
                'report_needless_disables')

        self.add_int_args('max_warning')

        self.add_str_args('config', 'config_basedir', 'ignore_path',
                'ignore_pattern', 'syntax', 'custom_syntax', 'cache_location',
                'formatter', 'custom_formatter')


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=' '.join(self.file_in),
        ))


def configure(conf):
    bin_path = 'node_modules/stylelint/bin/stylelint.js'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
        conf.env['%s_BIN' % tool_name.upper()] = bin_path
    else:
        conf.end_msg('not found', color='YELLOW')



"""
UglifyJS is a JavaScript compressor/minifier written in JavaScript. It also
contains tools that allow one to automate working with Javascript code.

Options:

    * source_map : str, None
                 : Specify an output file where to generate source map

    * source_map_root : str, None
                      : The path to the original source to be included in the
                      : source map

    * source_map_url : str, None
                     : The path to the source map to be added in
                     : //# sourceMappingURL.
                     : Defaults to the value passed with --source-map

    * source_map_include_sources : bool, None
                                 : Pass this flag if you want to include the
                                 : content of source files in the source map as
                                 : sourcesContent property

    * in_source_map : str, None
                    : Input source map, useful if you're compressing JS
                    : that was generated from some other original code

    * screw_ie8 : bool, None
                : Pass this flag if you don't care about full compliance with
                : Internet Explorer 6-8 quirks (by default UglifyJS will try to
                : be IE-proof)

    * expr : bool, None
           : Parse a single expression, rather than a program (for parsing JSON)

    * prefix : str, None
             : Skip prefix for original filenames that appear in source maps.
             : For example --prefix 3 will drop 3 directories from file names
             : and ensure they are relative paths. You can also specify --prefix
             : relative, which will make UglifyJS figure out itself the relative
             : paths between original sources, the source map and the output
             : file

    * beautify : str, None
               : Beautify output/specify output options

    * mangle : list, None
             : Mangle names/pass mangler options

    * reserved : list, None
               : Reserved names to exclude from mangling

    * compress : str, None
               : Enable compressor/pass compressor options. Pass options like
               : --compress hoist_vars=false,if_return=false.
               : Use --compress with no argument to use the default compression
               : options

    * define : str, None
             : Global definitions

    * enclose : str, None
              : Embed everything in a big function, with a configurable
              : parameter/argument list

    * comments : str, None
               : Preserve copyright comments in the output.
               : By default this works like Google Closure, keeping JSDoc-style
               : comments that contain "@license" or "@preserve". You can
               : optionally pass one of the following arguments to this flag:
               :  - "all" to keep all comments
               :  - a valid JS regexp (needs to start with a slash) to keep
               :    only comments that match.
               : Note that currently not *all* comments can be kept when
               : compression is on, because of dead code removal or cascading
               : statements into sequences

    * preamble : str, None
               : Preamble to prepend to the output. You can use this to insert
               : a comment, for example for licensing information. This will
               : not be parsed, but the source map will adjust for its presence

    * stats : bool, None
            : Display operations run time on STDERR

    * acorn : bool, None
            : Use Acorn for parsing

    * spidermonkey : bool, None
                   : Assume input files are SpiderMonkey AST format (as JSON)

    * self : bool, None
           : Build itself (UglifyJS2) as a library (implies --wrap=UglifyJS
           : --export-all)

    * wrap : str, None
           : Embed everything in a big function, making the "exports" and
           : "global" variables available. You need to pass an argument to this
           : option to specify the name that your model will take when included
           : in, say, a browser

    * export_all : bool, None
                 : Only used when --wrap, this tells UglifyJS to add code to
                 : automatically export all globals

    * lint : bool, None
           : Display some scope warnings

    * verbose : bool, None
              : Verbose

    * noerr : bool, None
            : Don't throw an error for unknown options in --compress,
            : --beautify, --mangle


Requirements:

    * node.js
    * uglify-js
      to install, run `npm install --save-dev uglify-js`

"""
import os
from shutil import copyfile, Error
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    conf = {
        '_replace_patterns_': ((r'\.js$', '.min.js'),),
    }

    def prepare(self):
        cfg = self.conf
        args = self.args

        self.add_bool_args('source_map_include_sources', 'screw_ie8', 'expr',
                'stats', 'acorn', 'spidermonkey', 'self', 'export_all', 'lint',
                'verbose', 'noerr')

        self.add_path_args('source_map_root', 'in_source_map')

        self.add_str_args('source_map', 'source_map_url', 'prefix', 'beautify',
                'reserved', 'define', 'enclose', 'preamble', 'wrap')

        if cfg.get('in_source_map', None):
            args.append("--in-source-map='%s'" % cfg['in_source_map'])

        for config in ('mangle', 'compress', 'comments'):
            if not config in cfg:
                continue

            c = cfg[config]
            if c:
                args.append("--%s='%s'" % (config, c))
            else:
                args.append('--' + config)


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        if not self.is_production():
            try:
                copyfile(self.file_in[0], self.file_out[0])
                return 0
            except (IOError, Error):
                self.bld.fatal('cannot copy file to ' + self.file_out[0])
            return -1

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} -o {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/uglify-js/bin/uglifyjs'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('uglifyjs')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Precompile handlebars templates.

Options:

    * amd          : bool, None, exports amd style (require.js)
    * commonjs     : bool, None, exports CommonJS style, path to Handlebars
                     module [default: null]
    * handlebarpath: str, None, path to handlebar.js (only valid for amd-style)
                     [default: ""]
    * known        : list, [], known helpers
    * knownOnly    : bool, None, known helpers only
    * minimize     : bool, None, minimize output
    * namespace    : str, None, template namespace
                     [default: 'Handlebars.templates']
    * simple       : bool, None, output template function only
    * root         : str, None, template root, base value that will be stripped
                     from template names
    * partial      : bool, None, compiling a partial template
    * data         : hash, None, include data when compiling
    * extension    : str, None, template extension [default: 'handlebars']
    * bom          : bool, None, removes the BOM (Byte Order Mark) from the
                     beginning of the templates

Requirements:

    * handlebars
      to install, edit package.json, run `npm install`
    * node.js

"""
import os
from json import dumps as json_dump
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.handlebars$', '.js'),),
    }
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        c = cfg.get('amd')
        if c:
            args.append('--amd')

        c = cfg.get('commonjs')
        if c:
            args.append('--commonjs')

        c = cfg.get('handlebarpath')
        if c:
            args.append("--handlebarPath='%s'" % c)

        for handler in make_list(cfg.get('known')):
            args.append("--known='%s'" % handler)

        c = cfg.get('known_only')
        if c:
            args.append('--knownOnly')

        c = cfg.get('minimize')
        if c:
            args.append('--min')

        c = cfg.get('namespace')
        if c:
            args.append("--namespace='%s'" % c)

        c = cfg.get('simple')
        if c:
            args.append('--simple')

        c = cfg.get('root')
        if c:
            args.append("--root='%s'" % c)

        c = cfg.get('partial')
        if c:
            args.append('--partial')

        c = cfg.get('data')
        if c:
            args.append("--data='%s'" % json_dump(c))

        c = cfg.get('extension')
        if c:
            args.append("--extension='%s'" % c)

        c = cfg.get('bom')
        if c:
            args.append('--bom')


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} -f {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0],
        ))


def configure(conf):
    bin_path = 'node_modules/handlebars/bin/handlebars'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('handlebars')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



import hashlib
import mmap
import os
import re
from ..core.group import Group
from .collections_utils import make_list

def get_filehash(filename):
    if not os.path.exists(filename):
        return None
    hasher = hashlib.sha1()
    with open(filename, 'rb') as f:
        with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as mm:
            hasher.update(mm)
    return hasher.digest()


def get_source_files(conf, bld):
    """Collect raw file inputs."""
    groups = {}
    constant_regex = re.compile(r'^[A-Z_]+$')

    def parse_group(group_name, config, level):
        groups['_%s' % level] = group_name

        options = config.pop('options', {})
        if group_is_leaf(config, options):
            group_files = make_list(config.get('raw_file_in')) +\
                    make_list(config.get('raw_depend_in'))

            for f in group_files:
                f = f.format(**groups)
                if os.path.isabs(f):
                    yield f
                else:
                    yield os.path.join(bld.top_dir, f)

            return

        for subgroup in config:
            for f in parse_group(subgroup, config[subgroup], level + 1):
                yield f

    for group in conf:
        if constant_regex.match(group):
            continue

        for f in parse_group(group, conf[group], 1):
            yield f


def group_is_leaf(group, options):
    """The lowests in the group tree are the tools."""

    return any(x in group for x in ('file_in', 'raw_file_in', 'file_out',
            'raw_file_out', 'depend_in', 'raw_depend_in', 'extra_out',
            'raw_extra_out', 'rule_in'))\
            or options.get('_no_io_', False)


def prepare_targets(conf, bld):
    """Create waf targets from predefined file input categories."""

    groups = {}
    constant_regex = re.compile(r'^[A-Z_]+$')

    def _add_raw_files(raw_file_list, file_list, pattern):
        for f in raw_file_list:
            f = f.format(**pattern)
            # because realpath() will remove the last path separator,
            # we need it to identify a directory
            is_dir = f.endswith(os.path.sep) or f.endswith('/')
            f = os.path.realpath(f)
            if is_dir and not f.endswith(os.path.sep):
                file_list.append(f + os.path.sep)
            elif '*' in f or '?' in f:
                for node in bld.root.ant_glob(f.lstrip('/')):
                    file_list.append(node.abspath())
            else:
                file_list.append(f)


    def _parse_input_listing(source_list, pattern):
        for f in source_list:
            f = f.format(**pattern)
            if f.startswith('@'):
                for x in groups[f[1:]].rule.files:
                    yield x
            elif not ('*' in f or '?' in f):
                yield f
            # expands wildcards (using ant_glob)
            elif os.path.isabs(f):
                for node in bld.root.ant_glob(f.lstrip('/')):
                    yield node.abspath()
            else:
                for node in bld.path.ant_glob(f):
                    yield node.relpath()


    def parse_group(group_name, config, level, parent_group):
        try:
            options = config.pop('options', {})
        except Exception as e:
            print((parent_group.get_name(), group_name, dict(config), level))
            raise e

        g = Group(group_name, parent_group, options)
        if parent_group is None:
            g.context = bld

        groups[g.get_name()] = g
        pattern = g.get_patterns()

        if group_is_leaf(config, options):
            original_file_in = make_list(config.get('file_in'))
            file_in = list(_parse_input_listing(original_file_in, pattern))
            _add_raw_files(make_list(config.get('raw_file_in')), file_in,
                    pattern)

            original_depend_in = make_list(config.get('depend_in'))
            depend_in = list(_parse_input_listing(original_depend_in, pattern))
            _add_raw_files(make_list(config.get('raw_depend_in')), depend_in,
                    pattern)

            original_file_out = make_list(config.get('file_out'))
            file_out = [x.format(**pattern) for x in original_file_out]
            _add_raw_files(make_list(config.get('raw_file_out')), file_out,
                    pattern)

            original_extra_out = make_list(config.get('extra_out'))
            extra_out = [x.format(**pattern) for x in original_extra_out]
            _add_raw_files(make_list(config.get('raw_extra_out')), extra_out,
                    pattern)

            rules_in = [x.format(**pattern) for x in make_list(
                    config.get('rule_in'))]

            for rule_in in rules_in:
                try:
                    token_names = bld._token_names[rule_in]
                    for f in token_names:
                        depend_in.append(f)
                except (KeyError, AttributeError):
                    print((parent_group.get_name(), group_name, dict(config),
                            level))

                    if hasattr(bld, '_token_names'):
                        bld.fatal("rule '%s' not found in: %s" % (rule_in,
                                ', '.join(bld._token_names.keys())))
                    else:
                        bld.fatal("rule '%s' not found" % rule_in +\
                                ", does it have *_in or *_out?")

            g(file_in=file_in, file_out=file_out, depend_in=depend_in,
                    extra_out=extra_out)
            return

        for subgroup in config:
            parse_group(subgroup, config[subgroup], level + 1, g)

    for group in conf:
        if constant_regex.match(group):
            continue
        parse_group(group, conf[group], 1, None)

    bld.task_gen_cache_names = groups



"""
convert existing TrueType/OpenType fonts to WOFF format (subject to
appropriate licensing)

Requirements:

    * sfnt2woff
      to install, download/compile from http://people.mozilla.org/~jkew/woff/

"""
import re
from pybuildtool import BaseTask

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.ttf$', '.woff'), (r'\.otf$', '.woff'))
    }
    name = tool_name

    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s only have one output' % tool_name.capitalize())

        executable = self.env['%s_BIN' % tool_name.upper()]
        ret = self.exec_command(
            '{exe} {arg} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
        ))

        if ret == 0:
            # success exit code
            move_executable = self.env['MV_BIN']
            converted_file = self.file_in[0]
            for (pat, rep) in self.conf['_replace_patterns_']:
                converted_file = re.sub(pat, rep, converted_file)
            ret = self.exec_command(
                '{exe} {in_} {out}'.format(
                exe=move_executable,
                in_=converted_file,
                out=self.file_out[0],
            ))

        return ret


def configure(conf):
    if not conf.env.MV_BIN:
        conf.env.MV_BIN = conf.find_program('mv')[0]

    bin_path = conf.find_program('sfnt2woff')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
Run shell commands.

Options:

    * work_dir : str, None
               : Change current directory before running shell command

    * command  : str, None
               : shell command

    * environ  : dict, None
               : Key value pair of shell environment.

    * fresh_environ : bool, False
                    : Starts with empty environment

"""
import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    name = tool_name
    workdir = None
    cmd = None
    environ = None
    fresh_environ = None

    def prepare(self):
        cfg = self.conf

        # Change current directory
        c = cfg.get('work_dir', None)
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(c + ' not found.')
        else:
            c = os.environ.get('ROOT_DIR')
            if c:
                self.workdir = c

        c = cfg.get('command')
        if c:
            self.cmd = c.format(**self.group.get_patterns())
        else:
            self.bld.fatal('command option is required.')

        # Shell environment
        self.environ = {}
        c = cfg.get('environ', None)
        if c:
            for key, value in c.items():
                self.environ[key] = value.format(**self.group.get_patterns())

        self.fresh_environ = cfg.get('fresh_environ', False)


    def perform(self):
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        if self.fresh_environ:
            environ = {k: v for k, v in os.environ.items() if k in ('PATH',)}
        else:
            environ = os.environ.copy()
        for key, value in self.environ.items():
            environ[key] = value

        kwargs = {
            'stdout': None,
            'stderr': None,
            'shell': True,
            'env': environ,
        }
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=self.cmd,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fonts_public.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='fonts_public.proto',
  package='google.fonts',
  serialized_pb=_b('\n\x12\x66onts_public.proto\x12\x0cgoogle.fonts\"\xae\x01\n\x0b\x46\x61milyProto\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x10\n\x08\x64\x65signer\x18\x02 \x02(\t\x12\x0f\n\x07license\x18\x03 \x02(\t\x12\x10\n\x08\x63\x61tegory\x18\x04 \x02(\t\x12\x12\n\ndate_added\x18\x05 \x02(\t\x12&\n\x05\x66onts\x18\x06 \x03(\x0b\x32\x17.google.fonts.FontProto\x12\x0f\n\x07\x61liases\x18\x07 \x03(\t\x12\x0f\n\x07subsets\x18\x08 \x03(\t\"\x8a\x01\n\tFontProto\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05style\x18\x02 \x02(\t\x12\x0e\n\x06weight\x18\x03 \x02(\x05\x12\x10\n\x08\x66ilename\x18\x04 \x02(\t\x12\x18\n\x10post_script_name\x18\x05 \x02(\t\x12\x11\n\tfull_name\x18\x06 \x02(\t\x12\x11\n\tcopyright\x18\x07 \x01(\tB%\n\x16\x63om.google.fonts.protoB\x0b\x46ontsPublic')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_FAMILYPROTO = _descriptor.Descriptor(
  name='FamilyProto',
  full_name='google.fonts.FamilyProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts.FamilyProto.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='designer', full_name='google.fonts.FamilyProto.designer', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='license', full_name='google.fonts.FamilyProto.license', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='category', full_name='google.fonts.FamilyProto.category', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='date_added', full_name='google.fonts.FamilyProto.date_added', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fonts', full_name='google.fonts.FamilyProto.fonts', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aliases', full_name='google.fonts.FamilyProto.aliases', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subsets', full_name='google.fonts.FamilyProto.subsets', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37,
  serialized_end=211,
)


_FONTPROTO = _descriptor.Descriptor(
  name='FontProto',
  full_name='google.fonts.FontProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='google.fonts.FontProto.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='style', full_name='google.fonts.FontProto.style', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weight', full_name='google.fonts.FontProto.weight', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='filename', full_name='google.fonts.FontProto.filename', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='post_script_name', full_name='google.fonts.FontProto.post_script_name', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='full_name', full_name='google.fonts.FontProto.full_name', index=5,
      number=6, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='copyright', full_name='google.fonts.FontProto.copyright', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=352,
)

_FAMILYPROTO.fields_by_name['fonts'].message_type = _FONTPROTO
DESCRIPTOR.message_types_by_name['FamilyProto'] = _FAMILYPROTO
DESCRIPTOR.message_types_by_name['FontProto'] = _FONTPROTO

FamilyProto = _reflection.GeneratedProtocolMessageType('FamilyProto', (_message.Message,), dict(
  DESCRIPTOR = _FAMILYPROTO,
  __module__ = 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts.FamilyProto)
  ))
_sym_db.RegisterMessage(FamilyProto)

FontProto = _reflection.GeneratedProtocolMessageType('FontProto', (_message.Message,), dict(
  DESCRIPTOR = _FONTPROTO,
  __module__ = 'fonts_public_pb2'
  # @@protoc_insertion_point(class_scope:google.fonts.FontProto)
  ))
_sym_db.RegisterMessage(FontProto)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\026com.google.fonts.protoB\013FontsPublic'))
# @@protoc_insertion_point(module_scope)



"""
webpack is a module bundler for modern JavaScript applications.

Options:

    * mode : str, None
           : Enable production optimizations or development hints.
           : values: development, production, none
           : If not set it'd be inferred from environment variables.

    * config_file : str, None
                  : path to the config file
                  : default: webpack.config.js or webpackfile.js

    * env : list, None
          : environment passed to the config, when it is a function

    * context : str, None
              : the root directory for resolving entry point and stats
              : default: The current directory

    * entry : dict<str, str>, None
            : the entry point

    * debug : bool, None
            : switch loaders to debug mode

    * devtool : list, None
              : enable devtool for better debugging experience
              : example: --devtool eval-cheap-module-source-map

    * progress : bool, None
               : print compilation progress in percentage

    * module_bind : dict<str, str>, None
                  : bind an extension to a loader

    * module-bind-post : dict<str, str>, None
    * module-bind-pre  : dict<str, str>, None

    * output_path : str, None
                  : the output path for compilation assets
                  : default: The current directory

    * output_filename : str, None
                      : the output filename of the bundle
                      : default: [name].js

    * output_chunk_filename : str, None
                            : the output filename for additional chunks
                            : default: filename with [id] instead of [name] or
                              [id] prefixed

    * output_source_map_filename : str, None
                                 : the output filename for the SourceMap

    * output_public_path : str, None
                         : the public path for the assets

    * output_jsonp_function : str, None
                            : the name of the jsonp function used for chunk
                            : loading

    * output_pathinfo : bool, None
                      : include a comment with the request for every
                      : dependency (require, import, etc.)

    * output_library : str, None
                     : expose the exports of the entry point as library

    * output_library_target : str, None
                            : the type for exposing the exports of the entry
                            : point as library

    * records_input_path : str, None
                         : path to the records file (reading)

    * records_output_path : str, None
                          : path to the records file (writing)

    * records_path : str, None
                   : path to the records file

    * define : dict<str, str>, None
             : define any free var in the bundle

    * target : str, None
             : the targeted execution environment

    * cache : bool, None
            : enable in memory caching
            : default: It's enabled by default when watching

    * watch_stdin : bool, None
                  : exit the process when stdin is closed

    * watch_aggregate_timeout : int, None
                              : timeout for gathering changes while watching

    * watch_poll : bool, None
                 : the polling interval for watching (also enable polling)

    * hot : bool, None
          : enables Hot Module Replacement

    * prefetch : list, None
               : prefetch this request (Example: --prefetch ./file.js)

    * provide : dict<str, str>, None
              : provide these modules as free vars in all modules
              : example: --provide jQuery=jquery

    * labeled_modules : bool, None
                      : enables labeled modules

    * plugin : list, None
             : load this plugin

    * bail : bool, None
           : abort the compilation on first error

    * profile : bool, None
              : profile the compilation and include information in stats

    * resolve_alias : dict<str, str>, None
                    : setup a module alias for resolving
                    : example: jquery-plugin=jquery.plugin

    * resolve_extensions : list, None
                         : setup extensions that should be used to resolve
                         : modules
                         : example: --resolve-extensions .es6 .js

    * resolve_loader_alias : dict<str, str>, None
                           : setup a loader alias for resolving

    * optimize_max_chunks : int, None
                          : try to keep the chunk count below a limit

    * optimize_min_chunk_size : int, None
                              : try to keep the chunk size above a limit

    * optimize_minimize : bool, None
                        : minimize javascript and switches loaders to minimizing

    * color : bool, None
            : enables/Disables colors on the console
            : default: (supports-color)

    * sort_modules_by : str, None
                      : sorts the modules list by property in module

    * sort_chunks_by : str, None
                     : sorts the chunks list by property in chunk

    * sort_assets_by : str, None
                     : sorts the assets list by property in asset

    * hide_modules : bool, None
                   : hides info about modules

    * display_exclude : list, None
                      : exclude modules in the output

    * display_modules : bool, None
                      : display even excluded modules in the output

    * display_max_modules : int, None
                          : sets the maximum number of visible modules in output

    * display_chunks : bool, None
                     : display chunks in the output

    * display_entrypoints : bool, None
                          : display entry points in the output

    * display_origins : bool, None
                      : display origins of chunks in the output

    * display_cached : bool, None
                     : display also cached modules in the output

    * display_cached_assets : bool, None
                            : display also cached assets in the output

    * display_reasons : bool, None
                      : display reasons about module inclusion in the output

    * display_depth : bool, None
                    : display distance from entry point for each module

    * display_used_exports : bool, None
                           : display information about used exports in modules
                           : (Tree Shaking)

    * display_provided_exports : bool, None
                               : display information about exports provided from
                               : modules

    * display_error_details : bool, None
                            : display details about errors

    * verbose : bool, None
              : show more details


Requirements:

    * webpack, webpack-cli
      to install, `npm install webpack webpack-cli`

"""
import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)

        c = cfg.get('mode', os.environ.get('NODE_ENV'))
        if not c:
            if self.bld.variant in ('prod', 'production'):
                c = 'production'
            else:
                c = 'development'
        args.append('--mode=' + c)

        self.add_bool_args('debug', 'verbose', 'progress', 'output_pathinfo',
                'cache', 'watch_stdin', 'watch_poll', 'hot', 'labeled_modules',
                'bail', 'profile', 'optimize_minimize', 'color', 'hide_modules',
                'display_modules', 'display_chunks', 'display_entrypoints',
                'display_origins', 'display_cached', 'display_cached_assets',
                'display_reasons', 'display_depth', 'display_used_exports',
                'display_provided_exports', 'display_error_details')

        self.add_dict_args('module_bind', 'module_bind_pre', 'module_bind_post',
                'define', 'provide', 'resolve_alias', 'resolve_loader_alias',
                opt_val_sep=' ')

        self.add_int_args('watch_aggregate_timeout', 'optimize_max_chunks',
                'optimize_min_chunk_size', 'display_max_modules')

        self.add_list_args_multi('devtool', 'plugin', 'display_exclude')

        self.add_list_args_multi('env', opt_val_sep='.')
        self.add_list_args_multi('resolve_extensions', opt_val_sep=' ')

        self.add_path_args('context', 'records_input_path')

        self.add_path_list_args_multi('prefetch')

        self.add_str_args('output_path', 'output_filename',
                'output_chunk_filename', 'output_source_map_filename',
                'output_public_path', 'output_jsonp_function', 'output_library',
                'output_library_target', 'records_output_path', 'records_path',
                'target', 'sort_modules_by', 'sort_chunks_by', 'sort_assets_by',
                )

        c = cfg.get('config_file')
        if c:
            args.append('--config=' + expand_resource(self.group, c))

        c = cfg.get('entry', {})
        for entry_name, entry_js_file in c.items():
            args.append('--%s=%s' % (entry_name, expand_resource(
                    self.group, entry_js_file)))


    def perform(self):
        if len(self.file_out) > 1:
            self.bld.fatal('%s at most produces one output' %\
                    tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            "{exe} {arg} {in_} {out}".format(
            exe=executable,
            arg=' '.join(self.args),
            in_=' '.join(self.file_in),
            out=' '.join(self.file_out),
        ),
        **kwargs)


def configure(conf):
    bin_path = 'node_modules/webpack-cli/bin/cli.js'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('webpack')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



import os
from waflib import Context, Errors # pylint:disable=import-error

class WatchContext(Context.Context):
    cmd = 'watch'
    fun = 'watch'
    variant = ''

    def __init__(self, **kw):
        super().__init__(**kw)

        self.top_dir = kw.get('top_dir', Context.top_dir)
        self.out_dir = kw.get('out_dir', Context.out_dir)

        if not(os.path.isabs(self.top_dir) and os.path.isabs(self.out_dir)):
            raise Errors.WafError('The project was not configured: ' +\
                    'run "waf configure" first!')
        self.path = self.srcnode = self.root.find_dir(self.top_dir)
        self.bldnode = self.root.make_node(self.variant_dir)


    def get_variant_dir(self):
        if not self.variant:
            return self.out_dir
        return os.path.join(self.out_dir, self.variant)

    variant_dir = property(get_variant_dir, None)



"""
Validate python files.

Options:

    * work_dir         : str,      None,  change current directory before
                                          running pylint
    * error_only       : bool,     False,  only check for errors
    * config_file      : str,      None,  pylint configuration file
    * plugins          : list:str, [] ,   plugins to load (ex. pylint_django)
    * reporter         : str,      None,  custom reporter
    * full_report      : bool,     False, full report or only the messages

Requirements:

    * pylint
      to install, run `pip install pylint`

"""
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_noop_retcodes_': [4],
    }
    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Change current directory, before running pylint, helps module imports
        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        # Specify a configuration file
        c = cfg.get('config_file')
        if c:
            args.append("--rcfile='%s'" % expand_resource(self.group, c))

        # Set the output format. Available formats are text,
        # parseable, colorized, msvs (visual studio) and html.
        # You can also give a reporter class, eg
        # mypackage.mymodule.MyReporterClass. [current: text]
        c = cfg.get('reporter', None)
        if c:
            args.append("--output-format='%s'" % c)

        # In error mode, checkers without error messages are
        # disabled and for others, only the ERROR messages are
        # displayed, and no reports are done by default
        c = cfg.get('error_only', False)
        if c:
            args.append('--errors-only')

        # Tells whether to display a full report or only the
        # messages [current: yes]
        c = cfg.get('full_report', False)
        if c:
            args.append('--reports=y')
        else:
            args.append('--reports=n')

        # Plugins
        plugins = make_list(cfg.get('plugins'))
        if plugins:
            args.append('--load-plugins=%s' % ','.join(plugins))


    def perform(self):
        if not self.file_in:
            self.bld.fatal('%s for %s needs input' % (tool_name.capitalize(),
                    self.token_out[0]))

        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    conf.env['%s_BIN' % tool_name.upper()] = conf.find_program(tool_name)[0]



def data_merge(a, b):
    """
    Merges b into a and return merged result.

    see http://stackoverflow.com/a/15836901

    NOTE: tuples and arbitrary objects are not handled as it is totally
    ambiguous what should happen.
    """
    if b is None:
        return a
    key = None
    try:
        #if a is None or isinstance(a, str) or isinstance(a, unicode) or
        #isinstance(a, int) or isinstance(a, long) or isinstance(a, float):

        if a is None or isinstance(a, (str, int, float)):
            # border case for first run or if a is a primitive
            a = b
        elif isinstance(a, list):
            # lists can be only appended
            if isinstance(b, list):
                # merge lists
                for c in b:
                    if not c in a:
                        a.append(c)
            elif not b in a:
                # append to list
                a.append(b)
        elif isinstance(a, dict):
            # dicts must be merged
            if isinstance(b, dict):
                for key in b:
                    if key in a:
                        a[key] = data_merge(a[key], b[key])
                    else:
                        a[key] = b[key]
            else:
                raise Exception('Cannot merge non-dict "%s" into dict "%s"' %\
                (b, a))

        else:
            raise Exception('NOT IMPLEMENTED "%s" into "%s"' % (b, a))
    except TypeError as e:
        raise Exception(
                'TypeError "%s" in key "%s" when merging "%s" into "%s"' %\
                (e, key, b, a)) from e

    return a


def is_non_string_iterable(data):
    """Check if data was iterable but not a string."""
    # http://stackoverflow.com/a/17222092
    try:
        if isinstance(data, (unicode, str)):
            return False
    except NameError:
        pass
    if isinstance(data, bytes):
        return False
    try:
        iter(data)
    except TypeError:
        return False
    try:
        hasattr(None, data)
    except TypeError:
        return True
    return False


def make_list(items, nodict=False):
    """If items was not a list, create a list with it as a member."""
    if items is None:
        return []
    if not is_non_string_iterable(items):
        return [items]
    if nodict and isinstance(items, dict):
        return [items]

    return items



import os
from waflib.Logs import debug # pylint:disable=import-error
from .rule import Rule
from ..misc.collections_utils import data_merge

class Group():

    name = None
    conf = None
    env = None
    context = None
    group = None

    level = 1
    rule = None

    def __init__(self, name, group, config):
        self.name = name
        self.conf = {}
        if group is not None:
            self.group = group
            self.level = group.level + 1
            self.context = group.context
            data_merge(self.conf, group.conf)
        data_merge(self.conf, config)


    def get_name(self):
        names = [self.name]
        parent = self.group
        while parent:
            names.append(parent.name)
            parent = parent.group
        return '/'.join(reversed(names))


    def get_patterns(self):
        patterns = {}
        parent = self
        while parent:
            patterns['_%s' % parent.level] = parent.name
            parent = parent.group
        return patterns


    def __enter__(self):
        return self


    def __exit__(self, type_, value, traceback):
        pass


    def __call__(self, file_in, file_out, depend_in, extra_out):
        bld = self.group.context
        try:
            task_class = bld.tools[self.name].Task
        except KeyError:
            bld.fatal('Unknown tool: ' + self.name)

        conf = {}
        data_merge(conf, self.conf)
        data_merge(conf, task_class.conf)

        self.rule = Rule(self, conf, file_in, file_out, depend_in, extra_out)
        for r in self.rule.rules:
            task = task_class(self.group, conf, env=bld.env)
            task_uid = task._id

            for f in r.get('file_in', []):
                if os.path.isabs(f):
                    if not os.path.exists(f):
                        continue
                    node = bld.root.find_resource(f.lstrip('/'))
                else:
                    node = bld.path.find_resource(f)
                if node is None:
                    bld.fatal('Source file "%s" does not exist' % f)

                debug('%s:%s: %s', 'input', 'file_in', str(node))
                task.set_inputs(node)

            for f in r.get('depend_in', []):
                if os.path.isabs(f):
                    node = bld.root.find_or_declare(f)
                else:
                    node = bld.path.find_or_declare(f)
                if node is None:
                    bld.fatal('"%s" does not exists' % f)
                setattr(node, 'is_virtual_in_' + task_uid, True)

                debug('%s:%s: %s', 'input', 'depend_in', str(node))
                task.set_inputs(node)

            for f in r.get('file_out', []):
                if f.startswith(os.path.sep):
                    # create outside files
                    f_dir = os.path.dirname(f)
                    try:
                        os.makedirs(f_dir)
                    except OSError:
                    #except FileExistsError:
                        pass
                    d_node = bld.root.find_dir(f_dir)
                    node = d_node.make_node(os.path.basename(f))
                else:
                    node = bld.path.find_or_declare(f)

                debug('%s:%s: %s', 'output', 'file_out', str(node))
                task.set_outputs(node)

            for f in r.get('extra_out', []):
                if f.startswith(os.path.sep):
                    # create outside files
                    f_dir = os.path.dirname(f)
                    try:
                        os.makedirs(f_dir)
                    except OSError:
                    #except FileExistsError:
                        pass
                    d_node = bld.root.find_dir(f_dir)
                    node = d_node.make_node(os.path.basename(f))
                else:
                    node = bld.path.find_or_declare(f)
                setattr(node, 'is_virtual_out_' + task_uid, True)

                debug('%s:%s: %s', 'output', 'extra_out', str(node))
                task.set_outputs(node)

            bld.add_to_group(task)
        return self.rule



"""
Produce rst files for in code files documentation.

Options:

    * work_dir    : str, None
                    Change working directory
    * output_dir  : str, None, required
                    Output directory
    * project_dir : str, None, required
                    Location of python modules
    * exclude     : list, []
                    Skip the documentation in this files

Requirements:

    * sphinx
      to install, run `pip install sphinx`

"""
from pybuildtool import BaseTask, expand_resource, expand_wildcard, make_list

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    workdir = None

    def prepare(self):
        cfg = self.conf
        args = self.args

        args.append('--force')

        c = cfg.get('work_dir')
        if c:
            self.workdir = expand_resource(self.group, c)
            if self.workdir is None:
                self.bld.fatal(cfg['work_dir'] + ' not found.')

        output_dir = expand_resource(self.group, cfg['output_dir'])
        if output_dir is None:
            self.bld.fatal(cfg['output_dir'] + ' not found.')
        args.append('-o ' + output_dir)

        project_dir = expand_resource(self.group, cfg['project_dir'])
        if project_dir is None:
            self.bld.fatal(cfg['project_dir'] + ' not found.')
        args.append(project_dir)

        for fname in make_list(cfg.get('exclude')):
            exclude_files = expand_wildcard(self.group, fname)
            if exclude_files is None:
                self.bld.fatal(fname + ' not found.')
            args.extend(exclude_files)


    def perform(self):
        if self.file_in:
            self.bld.fatal('%s takes no input' % tool_name.capitalize())
        if self.file_out:
            self.bld.fatal('%s produces no output' % tool_name.capitalize())

        kwargs = {}
        if self.workdir is not None:
            kwargs['cwd'] = self.workdir

        executable = self.env['SPHINX_APIDOC_BIN']
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
                exe=executable,
                arg=' '.join(self.args),
                in_=' '.join(self.file_in),
            ),
            **kwargs)


def configure(conf):
    conf.env['SPHINX_APIDOC_BIN'] = conf.find_program('sphinx-apidoc')[0]



"""
A command line tool for running JavaScript scripts that use the AMD API for
declaring and using JavaScript modules and regular JavaScript script files.

Options:

    * work_dir     : str, None, change current directory before run
    * config_file  : str, None, r.js build file

Requirements:

    * node.js
    * requirejs
      to install, run `npm install --save-dev requirejs`

"""

import os
from pybuildtool import BaseTask, expand_resource

tool_name = __name__

class Task(BaseTask):

    name = tool_name
    prefix = []

    def prepare(self):
        cfg = self.conf

        # Change current directory
        c = cfg.get('work_dir', None)
        if c:
            self.prefix.append('cd %s;' \
                % expand_resource(self.group, c))

        self.args.append('-o %s' % expand_resource(self.group,
            cfg['config_file']))


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{pre} {exe} {arg}'.format(
            exe=executable,
            pre=' '.join(self.prefix),
            arg=' '.join(self.args),
        ))


def configure(conf):
    bin_path = 'node_modules/requirejs/bin/r.js'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
        conf.env['%s_BIN' % tool_name.upper()] = bin_path
    else:
        conf.end_msg('not found', color='YELLOW')



"""
Validate javascript files.

Options:

    * config_file      : str, None, jshint configuration file
    * reporter         : str, None, custom reporter
    * ignore_files     : list, [],  excludes files matching pattern
    * ignore_list_file : str, None, jshintignore file

Requirements:

    * node.js
    * jshint
      to install, run `node install --save-dev jshint`

"""

import os
from pybuildtool import BaseTask, make_list

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_source_grouped_': True,
    }
    name = tool_name

    def prepare(self):
        bld = self.group.context
        cfg = self.conf
        args = self.args

        # Custom configuration file
        if cfg.get('config_file', None):
            args.append("--config='%s'" % bld.path.find_resource(
                cfg['config_file']).abspath())

        # Custom reporter (<PATH>|jslint|checkstyle)
        if cfg.get('reporter', None):
            args.append('--reporter=%s' % cfg['reporter'])

        # Exclude files matching the given filename pattern
        # (same as .jshintignore)
        exclude_files = make_list(cfg.get('ignore_files'))
        for exclude_file in exclude_files:
            args.append('--exclude=%s' % exclude_file)

        # Pass in custom jshintignore file path
        if cfg.get('ignore_list_file', None):
            args.append("--exclude-path='%s'" % bld.path.find_resource(
                cfg['ignore_list_file']).abspath())


    def perform(self):
        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=' '.join(self.file_in),
        ))


def configure(conf):
    bin_path = 'node_modules/jshint/bin/jshint'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
    else:
        conf.end_msg('not found', color='YELLOW')
        bin_path = conf.find_program('jshint')[0]
    conf.env['%s_BIN' % tool_name.upper()] = bin_path



"""
A rich and powerful templating language for JavaScript.

Options:

    * force   : bool, False, force compilation to continue on error
    * filters : list, [], give the compiler a comma-delimited list of
                asynchronous filters, required for correctly generating code
    * name    : str, None, specify the template name when compiling a single
                file
    * basedir : str, None, if `basedir` is set, `name` will be ignored
                used to automatically generate `name`
    * include : list, None, include a file or folder which match the regex but
                would otherwise be excluded
                you can use this flag multiple times
                default: ["\\.html$", "\\.jinja$"]
    * exclude : list, None, exclude a file or folder which match the regex but
                would otherwise be included
                you can use this flag multiple times
                default: []

Requirements:

    * node.js
    * nunjucks
      to install, run `npm install --save-dev nunjucks`

"""
import os
from pybuildtool import BaseTask, expand_resource, make_list

tool_name = __name__

class Task(BaseTask):

    conf = {
        '_replace_patterns_': ((r'\.html', '.js'),),
    }
    name = tool_name

    def prepare(self):
        cfg = self.conf
        args = self.args

        # Change current directory
        c = cfg.get('force', False)
        if c:
            args.append('--force')

        c = make_list(cfg.get('filters'))
        if c:
            args.append('--filters="%s"' % ','.join(c))

        c_basedir = cfg.get('basedir', None)
        c = cfg.get('name', None)
        if c and c_basedir is not None:
            args.append('--name="%s"' % c)

        c = make_list(cfg.get('include'))
        for o in c:
            args.append('--include="%s"' % expand_resource(self.group, o))

        c = make_list(cfg.get('exclude'))
        for o in c:
            args.append('--exclude="%s"' % expand_resource(self.group, o))


    def perform(self):
        if len(self.file_in) != 1:
            self.bld.fatal('%s only need one input' % tool_name.capitalize())
        if len(self.file_out) != 1:
            self.bld.fatal('%s can only have one output' %\
                    tool_name.capitalize())

        file_in = self.file_in[0]
        basedir = self.conf.get('basedir', None)
        if basedir:
            basedir = expand_resource(self.group, basedir)
            if file_in.startswith(basedir):
                name = file_in[len(basedir) + 1:]
            else:
                name = os.path.basename(file_in)
            self.args.append('--name="%s"' % name)

        executable = self.env['%s_BIN' % tool_name.upper()]
        return self.exec_command(
            '{exe} {arg} {in_} > {out}'.format(
            exe=executable,
            arg=' '.join(self.args),
            in_=self.file_in[0],
            out=self.file_out[0]
        ))


def configure(conf):
    bin_path = 'node_modules/nunjucks/bin/precompile'
    conf.start_msg("Checking for program '%s'" % tool_name)
    if os.path.exists(bin_path):
        bin_path = os.path.realpath(bin_path)
        conf.end_msg(bin_path)
        conf.env['%s_BIN' % tool_name.upper()] = bin_path
    else:
        conf.end_msg('not found', color='YELLOW')



# make dictionary loaded by yaml has order of dictionary keys equal to how it
# was written
# see http://stackoverflow.com/q/13297744/319817
import yaml
import yaml.constructor

try:
    # included in standard lib from Python 2.7
    from collections import OrderedDict
except ImportError:
    # try importing the backported drop-in replacement
    # it's available on PyPI
    from ordereddict import OrderedDict

class OrderedDictYAMLLoader(yaml.Loader):
    """
    A YAML loader that loads mappings into ordered dictionaries.
    """

    def __init__(self, *args, **kwargs):
        yaml.Loader.__init__(self, *args, **kwargs)

        self.add_constructor('tag:yaml.org,2002:map',
                type(self).construct_yaml_map)

        self.add_constructor('tag:yaml.org,2002:omap',
                type(self).construct_yaml_map)

    def construct_yaml_map(self, node):
        data = OrderedDict()
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_mapping(self, node, deep=False):
        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(None, None,
                'expected a mapping node, but found %s' % node.id,
                node.start_mark)

        mapping = OrderedDict()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.constructor.ConstructorError(
                        'while constructing a mapping',
                        node.start_mark,
                        'found unacceptable key (%s)' % exc,
                        key_node.start_mark)

            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping
